// Nandakumar Edamana
// 2023-06-23, 2023-06-24

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#define DARK  1
#define LIGHT 0
#define MAXDIM 256

#define BENCHMARK_CYCLES 1000

// stride = width
typedef struct {
	int width;
	int height;
	char map[MAXDIM * MAXDIM];
} Bitmap;

void clear(Bitmap *bm)
{
	memset(bm->map, 0, MAXDIM * MAXDIM);
}

// Deliberately "inefficient" to mimic my past self
void drawchecker_nested_hell(Bitmap *bm)
{
// Disabled in favor of benchmarking
//	assert(bm->width / 8 == 0);
//	assert(bm->height / 8 == 0);

	for(int y = 0; y < bm->height / 8; ++y)
		for(int x = 0; x < bm->width / 8; ++x)
			for(int j = 0; j < 8; ++j)
				for(int i = 0; i < 8; ++i)
					bm->map[bm->width * ((y * 8) + j) + ((x * 8) + i)] =
						((x + y) % 2)? DARK: LIGHT;
}

// The only optimization I want to do is to move the decision making
void drawchecker_nested_hell_handoptim_ifout(Bitmap *bm)
{
// Disabled in favor of benchmarking
//	assert(bm->width / 8 == 0);
//	assert(bm->height / 8 == 0);

	for(int y = 0; y < bm->height / 8; ++y) {
		//int y8 = y * 8; // precalc like this actually makes things slower?
	
		for(int x = 0; x < bm->width / 8; ++x) {
			char color = ((x + y) % 2)? DARK: LIGHT;

			for(int j = 0; j < 8; ++j) {
				// These precalcs helped, though
				//int rowoffset = bm->width * ((y * 8) + j);
				//int checkoffset = rowoffset + (x * 8);
			
				for(int i = 0; i < 8; ++i)
					//bm->map[checkoffset + i] = color;
					bm->map[bm->width * ( (y * 8) + j) + (x * 8) + i] = color;
			}
		}
	}
}

void drawchecker_nested_hell_handoptim_noloop(Bitmap *bm)
{
// Disabled in favor of benchmarking
//	assert(bm->width / 8 == 0);
//	assert(bm->height / 8 == 0);

	for(int y = 0; y < bm->height / 8; ++y) {
		for(int x = 0; x < bm->width / 8; ++x) {
			// Assuming long is 8 bytes long and DARK is 1
			long color = ((x + y) % 2)? 0x0101010101010101: 0;

			for(int j = 0; j < 8; ++j) {
				*(long *)(bm->map + bm->width * ( (y * 8) + j) + (x * 8)) = color;
			}
		}
	}
}

void drawchecker_bitwise_long(Bitmap *bm)
{
	for(int y = 0; y < bm->height; ++y)
		for(int x = 0; x < bm->width; ++x)
			bm->map[bm->width * y + x] =
				(((y >> 3) & 1) ^ ((x >> 3) & 1))? DARK: LIGHT;
}

void drawchecker_bitwise_short(Bitmap *bm)
{
	for(int y = 0; y < bm->height; ++y)
		for(int x = 0; x < bm->width; ++x)
			bm->map[bm->width * y + x] =
				(((x >> 3) ^ (y >> 3)) & 1)? DARK: LIGHT;
}

void drawchecker_bitwise_short_noif_array(Bitmap *bm)
{
	char colors[2] = { LIGHT, DARK };

	for(int y = 0; y < bm->height; ++y)
		for(int x = 0; x < bm->width; ++x)
			bm->map[bm->width * y + x] =
				colors[((x >> 3) ^ (y >> 3)) & 1];
}

// Works only because LIGHT = 0 and DARK = 1
void drawchecker_bitwise_short_noif_direct(Bitmap *bm)
{
	for(int y = 0; y < bm->height; ++y)
		for(int x = 0; x < bm->width; ++x)
			bm->map[bm->width * y + x] = ((x >> 3) ^ (y >> 3)) & 1;
}

int write(Bitmap *bm, char *filename)
{
	FILE *fp = fopen(filename, "wb");
	if(!fp)
		return -1;
	
	fputs("P1\n", fp);
	fprintf(fp, "%d %d\n", bm->width, bm->height);
	for(int y = 0; y < bm->height; ++y) {
		for(int x = 0; x < bm->width; ++x) {
			fprintf(fp, "%d ", bm->map[bm->width * y + x]);
		}
		
		fputs("\n", fp);
	}
	
	fclose(fp);
	
	return 0;
}

void write_with_error(Bitmap *bm, char *filename, char *progname)
{
	int err = write(bm, filename);
	if(err) {
		perror(progname);
		exit(EXIT_FAILURE);
	}
}

int main(int argc, char *argv[])
{
	Bitmap bm;
	bm.width = MAXDIM;
	bm.height = MAXDIM; // Just to show that it works for non-square dims

	// Write using every algorithm so that I can verify the outputs
	// Run this manually to make sure all images are the same:
	// for i in /tmp/checker_*; do diff "$i" /tmp/checker_bitwise_long.ppm; done

	clear(&bm);
	drawchecker_nested_hell(&bm);
	write_with_error(&bm, "/tmp/checker_nested_hell.ppm", argv[0]);

	clear(&bm);
	drawchecker_nested_hell_handoptim_ifout(&bm);
	write_with_error(&bm,
		"/tmp/checker_nested_hell_handoptim_ifout.ppm", argv[0]);

	clear(&bm);
	drawchecker_nested_hell_handoptim_noloop(&bm);
	write_with_error(&bm,
		"/tmp/checker_nested_hell_handoptim_noloop.ppm", argv[0]);
	
	clear(&bm);
	drawchecker_bitwise_long(&bm);
	write_with_error(&bm, "/tmp/checker_bitwise_long.ppm", argv[0]);

	clear(&bm);
	drawchecker_bitwise_short(&bm);
	write_with_error(&bm, "/tmp/checker_bitwise_short.ppm", argv[0]);

	clear(&bm);
	drawchecker_bitwise_short_noif_array(&bm);
	write_with_error(&bm, "/tmp/checker_bitwise_short_noif_array.ppm", argv[0]);

	clear(&bm);
	drawchecker_bitwise_short_noif_direct(&bm);
	write_with_error(&bm, "/tmp/checker_bitwise_short_noif_direct.ppm", argv[0]);

	// Now do the benchmarking

	clock_t cbeg_nested = clock();
	for(int i = 0; i < BENCHMARK_CYCLES; ++i)
		drawchecker_nested_hell(&bm);
	clock_t cend_nested = clock();

	clock_t cbeg_nested_handoptim_ifout = clock();
	for(int i = 0; i < BENCHMARK_CYCLES; ++i)
		drawchecker_nested_hell_handoptim_ifout(&bm);
	clock_t cend_nested_handoptim_ifout = clock();

	clock_t cbeg_nested_handoptim_noloop = clock();
	for(int i = 0; i < BENCHMARK_CYCLES; ++i)
		drawchecker_nested_hell_handoptim_noloop(&bm);
	clock_t cend_nested_handoptim_noloop = clock();

	clock_t cbeg_bwl = clock();
	for(int i = 0; i < BENCHMARK_CYCLES; ++i)
		drawchecker_bitwise_long(&bm);
	clock_t cend_bwl = clock();

	clock_t cbeg_bws = clock();
	for(int i = 0; i < BENCHMARK_CYCLES; ++i)
		drawchecker_bitwise_short(&bm);
	clock_t cend_bws = clock();

	clock_t cbeg_bws_noif_array = clock();
	for(int i = 0; i < BENCHMARK_CYCLES; ++i)
		drawchecker_bitwise_short_noif_array(&bm);
	clock_t cend_bws_noif_array = clock();

	clock_t cbeg_bws_noif_direct = clock();
	for(int i = 0; i < BENCHMARK_CYCLES; ++i)
		drawchecker_bitwise_short_noif_direct(&bm);
	clock_t cend_bws_noif_direct = clock();

	printf("time taken for drawchecker_nested_hell(): %fs\n",
		(cend_nested - cbeg_nested) / (double) CLOCKS_PER_SEC);

	printf("time taken for drawchecker_nested_hell_handoptim_ifout(): %fs\n",
		(cend_nested_handoptim_ifout - cbeg_nested_handoptim_ifout) /
			(double) CLOCKS_PER_SEC);

	printf("time taken for drawchecker_nested_hell_handoptim_noloop(): %fs\n",
		(cend_nested_handoptim_noloop - cbeg_nested_handoptim_noloop) /
			(double) CLOCKS_PER_SEC);

	printf("time taken for drawchecker_bitwise_long(): %fs\n",
		(cend_bwl - cbeg_bwl) / (double) CLOCKS_PER_SEC);

	printf("time taken for drawchecker_bitwise_short(): %fs\n",
		(cend_bws - cbeg_bws) / (double) CLOCKS_PER_SEC);

	printf("time taken for drawchecker_bitwise_short_noif_array(): %fs\n",
		(cend_bws_noif_array - cbeg_bws_noif_array) / (double) CLOCKS_PER_SEC);

	printf("time taken for drawchecker_bitwise_short_noif_direct(): %fs\n",
		(cend_bws_noif_direct - cbeg_bws_noif_direct) / (double) CLOCKS_PER_SEC);

	return EXIT_SUCCESS;
}
